#  fsh - fast remote execution
#  Copyright (C) 1999-2001 by Per Cederqvist.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import errno
import getopt
import getpass
import os
import select
import socket
import stat
import string
import sys
import time

import fshlib
import fshconfig

class remote:
    def __init__(self, method, login, server, background_wanted):
        self.__background_wanted = background_wanted
	in_fd, self.w = os.pipe()
	self.r, out_fd = os.pipe()
	self.child = os.fork()
	if self.child == 0:
	    # child
            os.close(self.w)
            os.close(self.r)
	    cmd = [method]
	    if login != 0:
		cmd = cmd + ["-l", login]
	    cmd = cmd + [server, 'in.fshd']
	    os.dup2(in_fd, 0)
	    os.dup2(out_fd, 1)
	    os.close(in_fd)
	    os.close(out_fd)
	    os.execvp(method, cmd)
	    sys.exit(1)
	# parent
	os.close(in_fd)
	os.close(out_fd)
	fshlib.set_nonblocking(self.w)
	fshlib.set_nonblocking(self.r)
	self.send_queue = []
	self.receive_queue = []
        self.state = 0

    def send(self, data):
	if data != "":
	    self.send_queue.append(data)
	if self.w == -1:
	    self.send_queue = []

    def select_action(self, r, w, e):
	if self.w in w:
	    if fshlib.write(self.w, self.send_queue) == -1:
		os.close(self.w)
		self.w = -1
		print "End of file writing to remote system."

	if self.r in r:
	    if fshlib.read(self.r, self.receive_queue, 4096) == -1:
		print "Tunnel closed."
		os.close(self.r)
		self.r = -1

    def select_set(self):
	# Return a tuple of three lists of file descriptors: the read,
	# write and exception fd sets.
	r = []
	w = []
	if self.w != -1 and self.send_queue != []:
	    w.append(self.w)
	if self.r != -1:
	    r.append(self.r)
	return (r, w, [])

    def __readline(self):
	if len(self.receive_queue) == 0:
	    return None
        ix = string.find(self.receive_queue[0], "\n")
        if ix == -1 and len(self.receive_queue) > 1:
	    self.receive_queue = [string.join(self.receive_queue, "")]
	    ix = string.find(self.receive_queue[0], "\n")
	if ix == -1:
            return None
        res = self.receive_queue[0][:ix]
        self.receive_queue[0] = self.receive_queue[0][ix+1:]
        return res

    def poll_response(self):
        while self.state == 0:
            s = self.__readline()
            if s == None:
                return [None, None, None]
            if s == "fsh 1":
                if self.__background_wanted:
                    background()
                self.state = 1
		print "Connection established"
                break
        return fshlib.parse_line(self.receive_queue, 1)

    def is_closed(self):
	return self.r == -1 or self.w == -1

class client:
    def __init__(self, socket):
        self.socket = socket.fileno()
        self.socket_object = socket
        self.send_queue = []
        self.receive_queue = []
	self.pending_close = 0

    def select_set(self):
        r = []
        w = []
        e = []
        live = 0
        if self.socket != -1:
            r.append(self.socket)
            if self.send_queue != []:
                w.append(self.socket)
            live = 1
        return (r, w, e, live)

    def select_action(self, r, w, e):
        # Handle any possible read/write.
        # Return 0 normally, 1 if the connection was unexpectedly lost.
        if self.socket == -1:
            return 0
        if self.socket in r:
	    if fshlib.read(self.socket, self.receive_queue, 4096) == -1:
		self.socket_object.close()
		self.socket = -1
                return 1

        if self.socket in w:
	    if fshlib.write(self.socket, self.send_queue) == -1:
		self.socket_object.close()
		self.socket = -1
		return 1
	    if self.pending_close and self.send_queue == []:
		self.socket_object.close()
		self.socket = -1

        return 0
                
    def poll_command(self):
        # return a parsed command, if one is available
        return fshlib.parse_line(self.receive_queue, 0)

    def send(self, data):
	if data != "":
	    self.send_queue.append(data)
	if self.socket == -1:
	    self.send_queue = []

    def close(self):
	self.pending_close = 1

    def is_closed(self):
	return self.socket == -1

    def drained(self):
	return self.send_queue == []

class fshd:
    def __init__(self, server, method, login, use_l_flag, background_wanted,
                 timeout):
	# Connect to server using method.  Parse away leading
	# noice until we receive "^fsh ".  Check the version number.
        sockname, sockdir = fshlib.fshd_socket(server, method, login)
        try:
            os.mkdir(sockdir, 0700)
        except os.error, (eno, emsg):
            if eno == errno.EEXIST:
                status = os.lstat(sockdir)
                if status[stat.ST_UID] != os.getuid():
                    raise
                if not stat.S_ISDIR(status[stat.ST_MODE]):
                    raise
                os.chmod(sockdir, 0700)
            else:
                raise
	self.r = remote(method, use_l_flag and login, server,
                        background_wanted)
        self.sessions = {}
        self.listen = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	self.listen.setblocking(0)
        self.listen.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	try:
	    os.unlink(sockname)
	except os.error, (eno, emsg):
	    if eno != errno.ENOENT:
		raise
        self.listen.bind(sockname)
        self.listen.listen(3)
        self.__timeout = timeout
        self.__expire_time = None
        self.setup_timer()

    def setup_timer(self):
        if self.__expire_time == None and self.__timeout != None:
            self.__expire_time = time.time() + self.__timeout

    def cancel_timer(self):
        if self.__expire_time != None:
            self.__expire_time = None

    def timeout(self):
        if self.__expire_time == None:
            return 10
        else:
            left = self.__expire_time - time.time()
            if left < 0:
                left = 0
            if left > 60:
                left = 60
            return left

    def expired(self):
        return self.__expire_time != None and self.__expire_time <= time.time()

    def toploop(self):
        try:
            while 1:
		drained = self.r.is_closed()
                r, w, e = self.r.select_set()
                r.append(self.listen.fileno())
                live = 0
                for s in self.sessions.keys():
		    if drained and not self.sessions[s].drained():
			drained = 0
                    r1, w1, e1, live1 = self.sessions[s].select_set()
                    r = r + r1
                    w = w + w1
                    e = e + e1
                    live = live + live1
		if drained:
		    sys.exit(0)
                if live > 0:
                    self.cancel_timer()
                else:
                    self.setup_timer()
                    if self.expired():
                        sys.exit(0)
                r, w, e = select.select(r, w, e, self.timeout())
                self.r.select_action(r, w, e)
                if self.listen.fileno() in r:
                    sock, addr = self.listen.accept()
		    sock.setblocking(0)
		    s = 1
		    while (self.sessions.has_key(s) and \
			   not self.sessions[s].is_closed()):
			s = s + 1
                    print s, "start"
                    self.sessions[s] = client(sock)
                while 1:
                    cmd, s, data = self.r.poll_response()
                    if cmd == None:
                        break
                    elif cmd in ["stdout", "stderr", "exit", "signal-exit", \
				 "stdin-flow"]:
			self.sessions[s].send("%s %s\n" % (
			    cmd, fshlib.hollerith(data)))
		    elif cmd in ["eof-stdin", "eof-stdout", "eof-stderr"]:
			self.sessions[s].send("%s\n" % cmd)
		    elif cmd == "eos":
			self.sessions[s].send("eos\n")
			self.sessions[s].close()
			print s, "EOS"
		    else:
			print "Unknown command from in.fshd:", cmd
                for s in self.sessions.keys():
                    if self.sessions[s].select_action(r, w, e) == 1:
                        # fsh died, so kill this session.
                        self.r.send("eos %d\n" % s)
                    while 1:
                        cmd, data = self.sessions[s].poll_command()
                        if cmd == None:
                            break
                        if data == None:
                            self.r.send("%s %d\n" % (cmd, s))
                        else:
                            self.r.send("%s %d %s\n" % (
                                cmd, s, fshlib.hollerith(data)))
			    if cmd == "new":
				print s, "$", data
        except KeyboardInterrupt:
            for s in self.sessions.keys():
                self.sessions[s].send("eos\n")
            sys.exit(0)

def background():
    try:
	signal.signal(signal.SIGTTOU, signal.SIG_IGN)
    except:
	pass
    try:
	signal.signal(signal.SIGTTIN, signal.SIG_IGN)
    except:
	pass
    try:
	signal.signal(signal.SIGTSTP, signal.SIG_IGN)
    except:
	pass

    if os.fork() > 0:
	os._exit(0)	# parent

    try:
	os.setsid()
    except:
	pass

    for i in range(3):
    	try:
	    os.close(i)
	except os.error, (eno, emsg):
	    if eno != errno.EBADF:
		raise
    fd = os.open("/dev/null", os.O_RDWR)
    assert fd == 0
    fd = os.dup(0)
    assert fd == 1
    fd = os.dup(0)
    assert fd == 2

def open_fds():
    """Return a list of file descriptors that might be open.
    """

    try:
        import resource
    except ImportError:
        # Fall back to using a large value.
        return range(1024)
    return range(resource.getrlimit(resource.RLIMIT_NOFILE)[0] + 1)

def usage(ret):
    sys.stderr.write(
	"fshd: usage: fshd [options] host\n"
	"             fshd { -h | --help }\n"
	"             fshd { -V | --version }\n")
    if ret == 0:
        sys.stderr.write(
            "Options:\n"
            "  -b --background       Run in the background.\n"
            "  -r method             Use ``method'' (e.g. ``rsh'') instead "
            "of ssh.\n"
            "  -l login              Log in as user ``login''.\n"
            "  -T --timeout=timeout  Set idle timeout (in seconds); exit when "
            "no session\n                        has existed for this long. "
            "0 disables the timeout.\n")

    sys.exit(ret)

def main():
    method = "ssh"
    login = None
    use_l_flag = 0
    print_version = 0
    b_flag = 0
    timeout = fshconfig.default_fshd_timeout
    try:
	opts, args = getopt.getopt(sys.argv[1:], "hr:l:bVT:",
                                   ["version", "help", "background",
                                    "timeout="])
    except getopt.error, msg:
	sys.stderr.write(str(msg) + "\n")
	sys.exit(1)
    for opt, val in opts:
	if opt == "-r":
	    method = fshlib.shell_unquote(val)
	elif opt == "-l":
	    login = val
	    use_l_flag = 1
	elif opt == "-b" or opt == "--background":
	    b_flag = 1
	elif opt == "-V" or opt == "--version":
	    print_version = 1
	elif opt == "-h" or opt == "--help":
	    usage(0)
        elif opt == "-T" or opt == "--timeout":
            timeout = string.atof(val)
        else:
            assert 0
    if print_version:
	fshlib.print_version("fshd")
    if len(args) != 1:
	usage(1)
    if login == None:
	login = getpass.getuser()

    for i in open_fds():
        if i > 2:
            try:
                os.close(i)
            except os.error, (eno, emsg):
                if eno != errno.EBADF:
                    raise

    s = fshd(args[0], method, login, use_l_flag, b_flag, timeout)
    s.toploop()

if __name__ == '__main__':
    main()



#  fsh - fast remote execution
#  Copyright (C) 1999-2001 by Per Cederqvist.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import errno
import getopt
import getpass
import os
import select
import socket
import string
import sys
import time

import fshlib

class remote:
    def __init__(self, method, login, use_l_flag, server, command,
                 fshd_timeout):
        (sockname, unused_dir) = fshlib.fshd_socket(server, method, login)
        self.s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	try:
	    self.s.connect(sockname)
	except socket.error, (eno, emsg):
	    if eno != errno.ECONNREFUSED and eno != errno.ENOENT:
		raise
	    else:
		# Start an fshd in daemon mode.
		cmd = "fshd -b -r " + fshlib.shell_quote(method)
		if use_l_flag:
		    cmd = cmd + " -l " + login
                if fshd_timeout != None:
                    cmd = cmd + " --timeout='%s'" % fshd_timeout
		cmd = cmd + " " + server
		ret = os.system(cmd)
		if not (os.WIFEXITED(ret) and os.WEXITSTATUS(ret) == 0):
		    sys.stderr.write("fsh: Failed to start fshd\n")
		    sys.exit(1)

		# Repeatedly try to connect to it.  Giv up after a minute.
		for i in range(60):
		    time.sleep(1)
		    try:
			self.s.connect(sockname)
			break
		    except socket.error, (eno, emsg):
			if eno != errno.ECONNREFUSED and eno != errno.ENOENT:
			    raise
		else:
		    sys.stderr.write("fsh: Failed to connect to tunnel.\n")
		    sys.exit(1)
			
	self.s.setblocking(0)
        self.sock = self.s.fileno()
        self.send_queue = []
        self.receive_queue = []
	self.eos_seen = 0
        self.send("new %s\n" % fshlib.hollerith(command))
        
    def select_set(self):
        w = []
        if self.sock == -1:
            return [], [], []
        if self.send_queue != []:
            w.append(self.sock)
        return [self.sock], w, []

    def select_action(self, r, w, e):
        if self.sock in r:
	    if fshlib.read(self.sock, self.receive_queue, 4096) == -1:
                self.s.close()
                self.sock = -1

        if self.sock in w:
	    if fshlib.write(self.sock, self.send_queue) == -1:
		self.s.close()
		self.sock = -1

    def send(self, data):
	#os.write(debug, "enqueueing this: %s.\n" % data)
        self.send_queue.append(data)
        if self.sock == -1:
            self.send_queue = []

    def poll_response(self):
        [cmd, data] = fshlib.parse_line(self.receive_queue, 0)

	# If half a command got through when we lost the tunnel, remove it.
	if cmd == None and self.sock == -1:
	    self.receive_queue = []

	if cmd == "eos":
	    self.eos_seen = 1

        return [cmd, data]

    def close(self):
        self.s.close()
        self.sock = -1

    def drained(self):
        return self.receive_queue == [] and self.sock == -1

class local_side:
    def __init__(self):
        self.stdin_quota = fshlib.QUOTA
        self.stdout_quota = fshlib.QUOTA
        self.stderr_quota = fshlib.QUOTA
        self.stdin_counter = 0
        self.stdout_counter = 0
        self.stderr_counter = 0
        self.stdout_queue = []
        self.stderr_queue = []
        self.stdin_fd = 0
        self.stdout_fd = 1
        self.stderr_fd = 2
	fshlib.set_nonblocking(self.stdin_fd)
	fshlib.set_nonblocking(self.stdout_fd)
	fshlib.set_nonblocking(self.stderr_fd)
        self.epipe_seen = 0
	self.stdout_pending_close = 0
	self.stderr_pending_close = 0

    def select_set(self):
        r = []
        w = []
        if self.stdin_fd != -1 and self.stdin_counter < self.stdin_quota:
            r.append(self.stdin_fd)
        if self.stdout_fd != -1 and self.stdout_queue != []:
            w.append(self.stdout_fd)
        if self.stderr_fd != -1 and self.stderr_queue != []:
            w.append(self.stderr_fd)
        return r, w, []

    def select_action(self, r, w, e, srv):
        if self.stdin_fd in r:
            wanted = min(4096, self.stdin_quota - self.stdin_counter)
            assert(wanted > 0)
	    queue = []
	    if fshlib.read(self.stdin_fd, queue, wanted) == -1:
                os.close(self.stdin_fd)
                self.stdin_fd = -1
                srv.send("eof-stdin\n")
	    elif len(queue) > 0:
		assert(len(queue) == 1)
                srv.send("stdin %s\n" % fshlib.hollerith(queue[0]))
                self.stdin_counter = self.stdin_counter + len(queue[0])

        if self.stdout_fd in w:
	    sz = fshlib.write(self.stdout_fd, self.stdout_queue)
	    if sz == -1:
		srv.send("eof-stdout\n")
		os.close(self.stdout_fd)
		self.stdout_fd = -1
	        #os.write(debug, "epipe seen I\n")
		self.epipe_seen = 1
	    else:
		self.stdout_counter = self.stdout_counter + sz

	    #os.write(debug, "successful write of %d bytes.\n" % sz)
            if self.stdout_pending_close and self.stdout_queue == []:
		#os.write(debug, "closing stdout due to pend.\n")
                os.close(self.stdout_fd)
                self.stdout_fd = -1
	    #os.write(debug, "hmm.\n")
            if self.stdout_fd != -1 and \
               (self.stdout_quota - self.stdout_counter) < fshlib.QUOTA/2:

		#os.write(debug, "increasing quota.\n")
                self.stdout_quota = self.stdout_quota + fshlib.QUOTA
                srv.send("stdout-flow %s\n" % fshlib.hollerith(
                    self.stdout_quota))
		#os.write(debug, "increased quota.\n")

        if self.stderr_fd in w:
	    sz = fshlib.write(self.stderr_fd, self.stderr_queue)
	    if sz == -1:
		srv.send("eof-stderr\n")
		os.close(self.stderr_fd)
		self.stderr_fd = -1
		#os.write(debug, "epipe seen II\n")
		self.epipe_seen = 1
	    else:
		self.stderr_counter = self.stderr_counter + sz
            self.stderr_counter = self.stderr_counter + sz
            if self.stderr_pending_close and self.stderr_queue == []:
                os.close(self.stderr_fd)
                self.stderr_fd = -1
            if self.stderr_fd != -1 and \
               (self.stderr_quota - self.stderr_counter) < fshlib.QUOTA/2:

                self.stderr_quota = self.stderr_quota + fshlib.QUOTA
                srv.send("stderr-flow %s\n" % fshlib.hollerith(
                    self.stderr_quota))

    def stdout_send(self, data):
	if data != "":
	    self.stdout_queue.append(data)

    def stderr_send(self, data):
	if data != "":
	    self.stderr_queue.append(data)

    def eof_stdin(self):
        if self.stdin_fd != -1:
            os.close(self.stdin_fd)
            self.stdin_fd = -1

    def eof_stdout(self):
        self.stdout_pending_close = 1
	if self.stdout_queue == [] and self.stdout_fd != -1:
	    os.close(self.stdout_fd)
	    self.stdout_fd = -1

    def eof_stderr(self):
	#os.write(debug, "eof_stderr: start\n")
        self.stderr_pending_close = 1
	#os.write(debug, "eof_stderr: I\n")
	if self.stderr_queue == [] and self.stderr_fd != -1:
	    #os.write(debug, "eof_stderr: II (%d, %d)\n" % (debug, self.stderr_fd))
	    os.close(self.stderr_fd)
	    #os.write(debug, "eof_stderr: III\n")
	    self.stderr_fd = -1
	    #os.write(debug, "eof_stderr: IV\n")
	#os.write(debug, "eof_stderr: V\n")

    def stdin_flow(self, data):
        self.stdin_quota = data

    def drained(self):
        return self.stdout_fd == -1 and self.stderr_fd == -1

def docmd(method, login, use_l_flag, server, command, fshd_timeout):
    rem = remote(method, login, use_l_flag, server, command, fshd_timeout)
    loc = local_side()
    exitval = None
    tunnel_lost = 0
    while 1:
	#os.write(debug, "at top of main loop\n")
        while 1:
            cmd, data = rem.poll_response()
            if cmd == None:
                break
	    #os.write(debug, "got cmd: %s (%s)\n" % (cmd, data))
            if cmd == "stdout":
                loc.stdout_send(data)
            elif cmd == "stderr":
                loc.stderr_send(data)
            elif cmd == "eof-stdin":
                loc.eof_stdin()
            elif cmd == "eof-stdout":
                loc.eof_stdout()
            elif cmd == "eof-stderr":
                loc.eof_stderr()
            elif cmd == "exit":
                exitval = string.atoi(data)
            elif cmd == "signal-exit":
                # What do we do here?
                exitval = string.atoi(data)
            elif cmd == "eos":
		#os.write(debug, "eos: start\n")
                loc.eof_stdin()
		#os.write(debug, "eos: stdin closed\n")
                loc.eof_stdout()
		#os.write(debug, "eos: stdout closed\n")
                loc.eof_stderr()
		#os.write(debug, "eos: stderr closed\n")
                rem.close()
		#os.write(debug, "eos: rem closed\n")
            elif cmd == "stdin-flow":
                loc.stdin_flow(string.atoi(data))
            else:
                raise "syntax error from server:" + str(cmd) + ":" + str(data)

	#os.write(debug, "main loop: commands performed\n")

        if loc.drained() and rem.drained():
            if exitval != None:
                if loc.epipe_seen and exitval == 0:
		    #os.write(debug, "epipe apparently seen\n")
                    sys.exit(1)
                else:
		    #os.write(debug, "exiting with %d\n" % exitval)
                    sys.exit(exitval)
            else:
                raise "don't know what to exit with"

	if rem.drained() and not rem.eos_seen and not tunnel_lost:
	    # We lost contact in mid-session.
	    # Schedule a shutdown of the client, but allow any pending
	    # data to be transmitted.
	    loc.eof_stdin()
	    loc.eof_stdout()
	    loc.stderr_send("Connection to fsh tunnel lost.\n")
	    loc.eof_stderr()
	    exitval = 1
	    tunnel_lost = 1

        r0, w0, e0 = rem.select_set()
        r1, w1, e1 = loc.select_set()
	#os.write(debug, "main loop: entering select\n")
        r, w, e = select.select(r0 + r1, w0 + w1, e0 + e1, 30)
	#os.write(debug, "main loop: dealing with remote select\n")
        rem.select_action(r, w, e)
	#os.write(debug, "main loop: dealing with local select\n")
        loc.select_action(r, w, e, rem)
	#os.write(debug, "main loop: end of loop\n")
        
def usage(ret):
    sys.stderr.write(
	"fsh: usage: fsh [options] host command [args...]\n"
	"            fsh { -h | --help }\n"
	"            fsh { -V | --version }\n")
    if ret == 0:
        sys.stderr.write(
            "Options:\n"
            "  -r method             Use ``method'' (e.g. ``rsh'') instead "
            "of ssh.\n"
            "  -l login              Log in as user ``login''.\n"
            "  -T --timeout=timeout  Set idle timeout for fshd (in seconds); exit when "
            "no\n                        session has existed for this long. "
            "0 disables.\n")
    sys.exit(ret)

def main():
    method = "ssh"
    login = None
    use_l_flag = 0
    print_version = 0
    fshd_timeout = None

    # First, look for options before the host.
    try:
	opts, args = getopt.getopt(sys.argv[1:], "hr:l:VT:",
                                   ["version", "help", "timeout="])
    except getopt.error, msg:
	sys.stderr.write(str(msg) + "\n")
	sys.exit(1)

    # For historical reasons, we have to look for options after the
    # host name as well.  This is ugly, but this syntax is widely
    # used, and we aim for rsh compatibility.  We only allow the
    # "-l login" after the host name.
    if len(args) < 2:
        host = None
        opts2 = []
    else:
        try:
            host = args[0]
            opts2, args = getopt.getopt(args[1:], "l:")
        except getopt.error, msg:
            sys.stderr.write(str(msg) + "\n")
            sys.exit(1)
    

    for opt, val in opts + opts2:
	if opt == "-r":
	    method = val
	elif opt == "-l":
	    login = val
	    use_l_flag = 1
	elif opt == "-V" or opt == "--version":
	    print_version = 1
	elif opt == "-h" or opt == "--help":
	    usage(0)
        elif opt == "-T" or opt == "--timeout":
            fshd_timeout = val
        else:
            assert 0

    if print_version:
	fshlib.print_version("fsh")
    if len(args) < 1 or host == None:
	usage(1)
    if login == None:
	login = getpass.getuser()
    docmd(method, login, use_l_flag, host, string.join(args),
          fshd_timeout)

if __name__ == "__main__":
    #global debug
    #debug = os.dup(2)
    #debug = os.open("/dev/null", os.O_RDWR)
    #os.write(debug", "debug: writing to fd %d.\n" % debug)
    main()



#  fsh - fast remote execution
#  Copyright (C) 2001 by Per Cederqvist.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

# Python 2.2 moved some constants from FCNTL to other modules, and
# dropped FD_CLOEXEC completely.  The relevant constants are now
# accessed via this module, that tries to get the constants as
# portably as possible.

import os
import fcntl

# O_NONBLOCK used to be present in FCNTL, but in Python 2.2 the only
# way to get it is via the os module.
try:
    O_NONBLOCK = os.O_NONBLOCK
except:
    import FCNTL
    O_NONBLOCK = FCNTL.O_NONBLOCK

# F_GETFD is present in FCNTL, but in Python 2.2 that module is
# deprecated, so we try to use fcntl instead.
try:
    F_GETFD = fcntl.F_GETFD
except:
    import FCNTL
    F_GETFD = FCNTL.F_GETFD

# F_SETFD is present in FCNTL, but in Python 2.2 that module is
# deprecated, so we try to use fcntl instead.
try:
    F_SETFD = fcntl.F_SETFD
except:
    import FCNTL
    F_SETFD = FCNTL.F_SETFD

# FD_CLOEXEC used to be present in FCNTL, but in Python 2.2 it isn't
# available at all!  I have reported the inability to find FD_CLOEXEC
# in Python 2.2 as a bug.  See <URL:http://sourceforge.net/tracker/
# ?func=detail&aid=496171&group_id=5470&atid=105470>.
try:
    # Stop Python 2.2 from warning that we import a deprecated module.
    # But Python 1.5.2 doesn't have the warnings module, so be prepared
    # for the import statement to fail.
    try:
        import warnings
        warnings.filterwarnings(
            "ignore",
            "the FCNTL module is deprecated; please use fcntl",
            DeprecationWarning)
    except ImportError:
        pass

    import FCNTL
    FD_CLOEXEC = FCNTL.FD_CLOEXEC

except AttributeError:

    # The value of FD_CLOEXEC is 1 on Solaris 8, HP-UX 11, Unicos 9.0,
    # Digital UNIX 4.0, SunOS 4.1.1_U1, IRIX 6.5, Linux 2.0 and Linux
    # 2.4 with glibc 2.2.3.  It is 1 even on AIX 4.2, so I guess it is
    # a fairly universal constant.

    FD_CLOEXEC = 1

    # If you find a system where the above definition of FD_CLOEXEC is
    # wrong, and where Python has been ported, please report it as a
    # bug to
    #
    #     http://bugzilla.lysator.liu.se/enter_bug.cgi?product=fsh
    #
    # Remember to state the operating system you are using (in the
    # format returned by os.uname()), and the proper value of
    # FD_CLOEXEC.



#  fsh - fast remote execution
#  Copyright (C) 1999-2001 by Per Cederqvist.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import errno
import fcntl
import getopt
import os
import select
import signal
import string
import sys

import fshlib
import fshcompat

def set_cloexec_flag(fd):
    oldflags = fcntl.fcntl(fd, fshcompat.F_GETFD)
    fcntl.fcntl(fd, fshcompat.F_SETFD, oldflags | fshcompat.FD_CLOEXEC)

def sig_to_name(sig):
    if sig == signal.SIGHUP:
	return "HUP"
    if sig == signal.SIGINT:
	return "INT"
    if sig == signal.SIGQUIT:
	return "QUIT"
    if sig == signal.SIGILL:
	return "ILL"
    if sig == signal.SIGTRAP:
	return "TRAP"
    if sig == signal.SIGIOT:
	return "IOT"
    if sig == signal.SIGBUS:
	return "BUS"
    if sig == signal.SIGFPE:
	return "FPE"
    if sig == signal.SIGKILL:
	return "KILL"
    if sig == signal.SIGUSR1:
	return "USR1"
    if sig == signal.SIGSEGV:
	return "SEGV"
    if sig == signal.SIGUSR2:
	return "USR2"
    if sig == signal.SIGPIPE:
	return "PIPE"
    if sig == signal.SIGALRM:
	return "ALRM"
    if sig == signal.SIGTERM:
	return "TERM"
    if sig == signal.SIGCHLD:
	return "CHLD"
    if sig == signal.SIGCONT:
	return "CONT"
    if sig == signal.SIGSTOP:
	return "STOP"
    if sig == signal.SIGTSTP:
	return "TSTP"
    if sig == signal.SIGTTIN:
	return "TTIN"
    if sig == signal.SIGTTOU:
	return "TTOU"
    if sig == signal.SIGURG:
	return "URG"
    if sig == signal.SIGXCPU:
	return "XCPU"
    if sig == signal.SIGXFSZ:
	return "XFSZ"
    if sig == signal.SIGVTALRM:
	return "VTALRM"
    if sig == signal.SIGPROF:
	return "PROF"
    if sig == signal.SIGWINCH:
	return "WINCH"
    if sig == signal.SIGIO:
	return "IO"
    if sig == signal.SIGPWR:
	return "PWR"
    return str(sig)

def name_to_sig(sig):
    if sig == "HUP":
	return signal.SIGHUP
    if sig == "INT":
	return signal.SIGINT
    if sig == "QUIT":
	return signal.SIGQUIT
    if sig == "ILL":
	return signal.SIGILL
    if sig == "TRAP":
	return signal.SIGTRAP
    if sig == "IOT":
	return signal.SIGIOT
    if sig == "BUS":
	return signal.SIGBUS
    if sig == "FPE":
	return signal.SIGFPE
    if sig == "KILL":
	return signal.SIGKILL
    if sig == "USR1":
	return signal.SIGUSR1
    if sig == "SEGV":
	return signal.SIGSEGV
    if sig == "USR2":
	return signal.SIGUSR2
    if sig == "PIPE":
	return signal.SIGPIPE
    if sig == "ALRM":
	return signal.SIGALRM
    if sig == "TERM":
	return signal.SIGTERM
    if sig == "CHLD":
	return signal.SIGCHLD
    if sig == "CONT":
	return signal.SIGCONT
    if sig == "STOP":
	return signal.SIGSTOP
    if sig == "TSTP":
	return signal.SIGTSTP
    if sig == "TTIN":
	return signal.SIGTTIN
    if sig == "TTOU":
	return signal.SIGTTOU
    if sig == "URG":
	return signal.SIGURG
    if sig == "XCPU":
	return signal.SIGXCPU
    if sig == "XFSZ":
	return signal.SIGXFSZ
    if sig == "VTALRM":
	return signal.SIGVTALRM
    if sig == "PROF":
	return signal.SIGPROF
    if sig == "WINCH":
	return signal.SIGWINCH
    if sig == "IO":
	return signal.SIGIO
    if sig == "PWR":
	return signal.SIGPWR
    return signal.SIGINT

class session:
    def __init__(self, session_no, cmd):
	# Spawn a process.
	in_fd, self.stdin_fd = os.pipe()
	self.stdout_fd, out_fd = os.pipe()
	self.stderr_fd, err_fd = os.pipe()
	set_cloexec_flag(self.stdin_fd)
	set_cloexec_flag(self.stdout_fd)
	set_cloexec_flag(self.stderr_fd)
	self.child = os.fork()
	if self.child == 0:
	    # child
	    os.dup2(in_fd, 0)
	    os.dup2(out_fd, 1)
	    os.dup2(err_fd, 2)
	    os.close(in_fd)
	    os.close(out_fd)
	    os.close(err_fd)
	    shell = os.environ.get("SHELL", "/bin/sh")
	    shell_name = string.split(shell, "/")[-1]
	    os.execv(shell, [shell_name, "-c", cmd])
	    sys.exit(1)
	# parent
	os.close(in_fd)
	os.close(out_fd)
	os.close(err_fd)
	fshlib.set_nonblocking(self.stdin_fd)
	fshlib.set_nonblocking(self.stdout_fd)
	fshlib.set_nonblocking(self.stderr_fd)
	self.session_no = session_no
	self.stdin_queue = []
	self.stdin_pending_close = 0
        self.stdin_quota = fshlib.QUOTA
        self.stdout_quota = fshlib.QUOTA
        self.stderr_quota = fshlib.QUOTA
        self.stdin_counter = 0
        self.stdout_counter = 0
        self.stderr_counter = 0

    def stdin(self, data):
	# Queue DATA for later transmission to the process.
        assert(not self.stdin_pending_close)
	if data != "":
	    self.stdin_queue.append(data)
	if self.stdin_fd == -1:
	    self.stdin_queue = []

    def select_action(self, r, w, e):
	# See if there is anything to do.

        # Return if an end-of-file was received from the child process.
        # That causes the top loop to reap children faster for a short
        # while, on the assumption that children often dies at the
        # same time as the file descriptors are closed.
        res = 0

	if self.stdin_fd in w:
	    sz = fshlib.write(self.stdin_fd, self.stdin_queue)
	    if sz == -1:
		os.close(self.stdin_fd)
		self.stdin_fd = -1
		# sys.stderr.write("infshd reporting eof-stdin\n")
		print "eof-stdin", self.session_no
                res = 1
	    else:
		self.stdin_counter = self.stdin_counter + sz
            if self.stdin_pending_close and self.stdin_queue == []:
                self.eof_stdin()
            if self.stdin_fd != -1 and \
               (self.stdin_quota - self.stdin_counter) < fshlib.QUOTA/2:

                self.stdin_quota = self.stdin_quota + fshlib.QUOTA
                print "stdin-flow", self.session_no, \
                      fshlib.hollerith(self.stdin_quota)

	if self.stdout_fd in r:
            wanted = min(4096, self.stdout_quota - self.stdout_counter)
            assert(wanted > 0)
	    queue = []
	    if fshlib.read(self.stdout_fd, queue, wanted) == -1:
                # sys.stderr.write("infshd reporting eof-stdout\n")
		print "eof-stdout", self.session_no
		os.close(self.stdout_fd)
		self.stdout_fd = -1
                res = 1
	    elif len(queue) > 0:
		assert(len(queue) == 1)
                # sys.stderr.write("infshd reporting stdout data\n")
                self.stdout_counter = self.stdout_counter + len(queue[0])
		print "stdout", self.session_no, fshlib.hollerith(queue[0])

	if self.stderr_fd in r:
            wanted = min(4096, self.stderr_quota - self.stderr_counter)
            assert(wanted > 0)
	    queue = []
	    if fshlib.read(self.stderr_fd, queue, wanted) == -1:
                # sys.stderr.write("infshd reporting eof-stderr\n")
		print "eof-stderr", self.session_no
		os.close(self.stderr_fd)
		self.stderr_fd = -1
                res = 1
	    elif len(queue) > 0:
		assert(len(queue) == 1)
                # sys.stderr.write("infshd reporting stderr data\n")
                self.stderr_counter = self.stderr_counter + len(queue[0])
		print "stderr", self.session_no, fshlib.hollerith(queue[0])

        return res


    def select_set(self):
	# Return a tuple of three lists of file descriptors: the read,
	# write and exception fd sets.
	r = []
	w = []
	if self.stdin_fd != -1 and self.stdin_queue != []:
	    w.append(self.stdin_fd)
	if self.stdout_fd != -1 and self.stdout_quota > self.stdout_counter:
	    r.append(self.stdout_fd)
	if self.stderr_fd != -1 and self.stderr_quota > self.stderr_counter:
	    r.append(self.stderr_fd)
	return (r, w, [])

    def wait_poll(self):
	# Check if the process has died by calling waitpid.
	if self.child == -1:
	    return
	try:
	    pid, status = os.waitpid(self.child, os.WNOHANG)
	except os.error, (eno, emsg):
	    if eno == errno.EINTR:
		return
	    raise
	if pid == 0:
	    return
	self.child = -1
	if os.WIFEXITED(status):
            # sys.stderr.write("infshd reporting exit\n")
	    print "exit", self.session_no, fshlib.hollerith(
                os.WEXITSTATUS(status))
	elif os.WIFSIGNALED(status):
            # sys.stderr.write("infshd reporting signal-exit\n")
	    print "signal-exit", self.session_no,
	    print fshlib.hollerith(sig_to_name(os.WTERMSIG(status)))
	else:
            # sys.stderr.write("infshd reporting bad exit\n")
	    print "exit", self.session_no, fshlib.hollerith("unknown")

    def signal(self, signal_name):
	# Send the signal SIGNAL_NAME to the process.
	if self.child == -1:
	    return
	os.kill(self.child, name_to_sig(signal_name))

    def eof_stdin(self):
	# Close stdin of the process.
        self.stdin_pending_close = 1
	if self.stdin_fd != -1 and self.stdin_queue == []:
	    os.close(self.stdin_fd)
	    self.stdin_fd = -1

    def eof_stdout(self):
	# Close stdout of the process.
	if self.stdout_fd != -1:
	    os.close(self.stdout_fd)
	    self.stdout_fd = -1

    def eof_stderr(self):
	# Close stderr of the process.
	if self.stderr_fd != -1:
	    os.close(self.stderr_fd)
	    self.stderr_fd = -1

    def eos(self):
	# Kill the process and close everything.
	self.signal("KILL")
        if self.stdin_fd != -1:
            os.close(self.stdin_fd)
            self.stdin_fd = -1
	self.eof_stdout()
	self.eof_stderr()
        # sys.stderr.write("infshd reporting eos\n")
	print "eos", self.session_no

    def spontaneous_eos(self):
	# Check if it is time to spontaneously do an eos.
	# Return true if an eos was performed.
	if (self.stdout_fd == -1 and self.stderr_fd == -1 \
	    and self.child == -1):

	    self.eos()
	    return 1
	else:
	    return 0

    def stdout_flow(self, new_quota):
        self.stdout_quota = string.atoi(new_quota)

    def stderr_flow(self, new_quota):
        self.stderr_quota = string.atoi(new_quota)


class infshd:
    def __init__(self):
	self.sessions = {}
	self.commands = []
	set_cloexec_flag(0)
	set_cloexec_flag(1)
	set_cloexec_flag(2)
        # sys.stderr.write("infshd starting\n")
	print "fsh 1"

    def new(self, session_no, cmd):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].eos()
	self.sessions[session_no] = session(session_no, cmd)

    def stdin(self, session_no, data):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].stdin(data)

    def signal(self, session_no, data):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].signal(data)

    def eof_stdin(self, session_no):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].eof_stdin()

    def eof_stdout(self, session_no):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].eof_stdout()

    def eof_stderr(self, session_no):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].eof_stderr()

    def eos(self, session_no):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].eos()
	    del self.sessions[session_no]

    def stdout_flow(self, session_no, data):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].stdout_flow(data)

    def stderr_flow(self, session_no, data):
	if self.sessions.has_key(session_no):
	    self.sessions[session_no].stderr_flow(data)

    def toploop(self):
        timeout = 4
	while 1:
	    r = [0]
	    w = []
	    e = []
	    for s in self.sessions.keys():
		r1, w1, e1 = self.sessions[s].select_set()
		r = r + r1
		w = w + w1
		e = e + e1
            sys.stdout.flush()
            # sys.stderr.write("infshd entering select\n")

	    # Don't wait too long -- we want to reap the children once
	    # in a while.
	    r, w, e = select.select(r, w, e, timeout)
            if timeout < 4:
                timeout = 1.5 * timeout

            # sys.stderr.write("infshd exiting select\n")
	    for s in self.sessions.keys():
                if self.sessions[s].select_action(r, w, e):
                    timeout = 0.05
		self.sessions[s].wait_poll()
		if self.sessions[s].spontaneous_eos():
		    del self.sessions[s]
	    if 0 in r:
		data = os.read(0, 1024)
		if data == "":
		    # End of file.  Kill everything.
		    for s in self.sessions.keys():
			self.sessions[s].eos()
		    sys.exit(0)
		self.commands.append(data)
		self.parse_commands()

    parse_table = [
	["new",        1, new],
	["stdin",      1, stdin],
	["signal",     1, signal],
	["eof-stdin",  0, eof_stdin],
	["eof-stdout", 0, eof_stdout],
	["eof-stderr", 0, eof_stderr],
	["eos",        0, eos],
	["stdout-flow", 1, stdout_flow],
	["stderr-flow", 1, stderr_flow],
	]

    def do_one_command(self):

	[parsed_cmd, parsed_session, data] = fshlib.parse_line(self.commands,
							       1)

	if parsed_cmd == None:
	    return 0

	for [cmd, need_data, cb] in self.parse_table:
	    if cmd == parsed_cmd:
		if need_data:
		    if data == None:
                        # sys.stderr.write("infshd reporting syntax error\n")
			print "syntax-error 3"
			return 1
		    cb(self, parsed_session, data)
		else:
		    if data != None:
                        # sys.stderr.write("infshd reporting syntax error\n")
			print "syntax-error 4"
			return 1
		    cb(self, parsed_session)
		return 1
        # sys.stderr.write("infshd reporting syntax error\n")
	print "syntax-error 5"
	return 1

    def parse_commands(self):
	while self.do_one_command() == 1:
	    pass

def usage(ret):
    sys.stderr.write(
        "in.fshd: usage: in.fshd [ -V | --version | -h | --help ]\n")
    sys.exit(ret)

def main():
    print_version = 0
    try:
	opts, args = getopt.getopt(sys.argv[1:], "hV",
                                   ["version", "help"])
    except getopt.error, msg:
	sys.stderr.write(str(msg) + "\n")
	sys.exit(1)
    for opt, val in opts:
	if opt == "-V" or opt == "--version":
	    print_version = 1
	elif opt == "-h" or opt == "--help":
	    usage(0)
        else:
            assert 0
    if print_version:
	fshlib.print_version("in.fshd")
    if len(args) != 0:
	usage(1)
    
    i = infshd()
    i.toploop()

if __name__ == '__main__':
    main()
